from .. import usbhid


profile = {
    "name": "SteelSeries Rival 100 / SteelSeries Rival 105",
    "models": [
        {
            "name": "SteelSeries Rival 100",
            "vendor_id": 0x1038,
            "product_id": 0x1702,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Rival 100 (Dell China)",
            "vendor_id": 0x1038,
            "product_id": 0x170A,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Rival 100 Dota 2 Edition (retail)",
            "vendor_id": 0x1038,
            "product_id": 0x170B,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Rival 100 Dota 2 Edition (Lenovo)",
            "vendor_id": 0x1038,
            "product_id": 0x170C,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Rival 105",
            "vendor_id": 0x1038,
            "product_id": 0x1814,
            "endpoint": 0,
        },
    ],
    "settings": {
        "sensitivity1": {
            "label": "Sensibility preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x01],
            "value_type": "choice",
            "choices": {
                250: 0x08,
                500: 0x07,
                1000: 0x06,
                1250: 0x05,
                1500: 0x04,
                1750: 0x03,
                2000: 0x02,
                4000: 0x01,
            },
            "default": 1000,
        },
        "sensitivity2": {
            "label": "Sensibility preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x02],
            "value_type": "choice",
            "choices": {
                250: 0x08,
                500: 0x07,
                1000: 0x06,
                1250: 0x05,
                1500: 0x04,
                1750: 0x03,
                2000: 0x02,
                4000: 0x01,
            },
            "default": 2000,
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },
        "color": {
            "label": "LED color",
            "description": "Set the mouse LED color",
            "cli": ["-c", "--color"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x05, 0x00],
            "value_type": "rgbcolor",
            "default": "#FF1800",
        },
        "light_effect": {
            "label": "Light effect",
            "description": "Set the light effect",
            "cli": ["-e", "--light-effect"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x07, 0x00],
            "value_type": "choice",
            "choices": {
                "steady": 0x01,
                "breath": 0x03,
                1: 0x01,
                2: 0x02,
                3: 0x03,
                4: 0x04,
            },
            "default": "steady",
        },
        "btn6_mode": {
            "label": "Button 6 mode",
            "description": "Set the mode of the button under the wheel",
            "cli": ["-b", "--btn6-mode"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x0B],
            "value_type": "choice",
            "choices": {
                "dpi": 0x00,
                "os": 0x01,
            },
            "default": "dpi",
        },
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },
    "firmware_version": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x10, 0x00],
        "response_length": 2,
    },
}



"""
This module generates rivalcfg's CLI.
"""


import os
import sys
import types
import platform
import argparse

from . import handlers
from . import devices
from . import udev
from . import debug
from .version import VERSION


def normalize_cli_option_name(name):
    """Helper function to transform a setting name to a cli option.

    :param str name: The setting name.
    :rtype: str

    >>> normalize_cli_option_name("My_Test_Setting1")
    'my-test-setting1'
    """
    return name.lower().replace("_", "-")


class PrintSupportedDevicesAction(argparse.Action):
    """Print supported devices and exit."""

    def __call__(self, parser, namespace, value, option_string=None):
        for item in [getattr(devices, name) for name in dir(devices)]:
            if not isinstance(item, types.ModuleType):
                continue
            if not hasattr(item, "profile"):
                continue
            print("%s:" % item.profile["name"])
            print()
            for model in item.profile["models"]:
                print(
                    "  %04x:%04x | %s"
                    % (
                        model["vendor_id"],
                        model["product_id"],
                        model["name"],
                    )
                )
            print()
        sys.exit(0)


class UpdateUdevRulesAction(argparse.Action):
    """Updates udev rules and exit."""

    def __call__(self, parser, namespace, value, option_string=None):
        if platform.system() != "Linux":
            print("E: The --update-udev option can only be used on Linux.")
            sys.exit(2)
        if os.getuid() != 0:
            print("E: You must run rivalcfg as root to use the --update-udev option.")
            sys.exit(2)
        udev.write_rules_file()
        udev.reload_rules()
        udev.trigger()
        sys.exit(0)


class PrintUdevRulesAction(argparse.Action):
    """Prints udev rules and exit."""

    def __call__(self, parser, namespace, value, option_string=None):
        print(udev.generate_rules())
        sys.exit(0)


class PrintDebugAction(argparse.Action):
    """Prints debug informations and exit."""

    def __call__(self, parser, namespace, value, option_string=None):
        print(debug.get_debug_info())
        sys.exit(0)


def add_main_cli(cli_parser):
    """Adds the main CLI options.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    """
    cli_parser.add_argument(
        "--list",
        help="List supported devices and exit",
        nargs=0,
        action=PrintSupportedDevicesAction,
    )

    cli_parser.add_argument(
        "--version",
        action="version",
        version=VERSION,
    )

    cli_parser.add_argument(
        "--no-save",
        help="Do not persist settings in the internal device memory",
        dest="SAVE",
        action="store_false",
        default=True,
    )

    cli_parser.add_argument(
        "--update-udev",
        help="Updates udev rules (Linux only, requires to be run as root)",
        nargs=0,
        action=UpdateUdevRulesAction,
    )

    cli_parser.add_argument(
        "--print-udev",
        help="Prints udev rules and exit",
        nargs=0,
        action=PrintUdevRulesAction,
    )

    cli_parser.add_argument(
        "--print-debug",
        help="Prints debug informations and exit",
        nargs=0,
        action=PrintDebugAction,
    )


def add_mouse_cli(cli_parser, mouse_profile):
    """Adds the CLI options for the given mouse profile.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param mouse_profile: One of the rivalcfg mouse profile (provided by
                            :func:`rivalcfg.devices.get_profile`).
    """
    cli_group = cli_parser.add_argument_group("%s Options" % mouse_profile["name"])

    for setting_name, setting_info in mouse_profile["settings"].items():
        handler = getattr(handlers, setting_info["value_type"])
        handler.add_cli_option(cli_group, setting_name, setting_info)

    cli_group.add_argument(
        "-r",
        "--reset",
        help="Reset all settings to their factory default",
        dest="RESET",
        action="store_true",
    )

    if "firmware_version" in mouse_profile:
        cli_group.add_argument(
            "--firmware-version",
            help="Print the firmware version of the mouse and exit",
            dest="FIRMWARE_VERSION",
            action="store_true",
        )

    if "battery_level" in mouse_profile:
        cli_group.add_argument(
            "--battery-level",
            help="Print the battery level of the mouse and exit",
            dest="BATTERY_LEVEL",
            action="store_true",
        )



"""
The "rgbgradientv2" type handles RGB color gradients. Simple RGB color can also
be used.

RGB gradient syntax example::

    rgbgradient(duration=1000; colors=0%: #ff0000, 33%: #00ff00, 66%: #0000ff)
    rgbgradient(colors=0%: red, 33%: lime, 66%: blue)

It supports both hexadecimal colors:

* ``#FF0000``
* ``FF0000``
* ``#F00``
* ``F00``

and named colors:

+------------+------------+-----------+-------------+
| ``white``  | ``red``    | ``lime``  | ``blue``    |
+------------+------------+-----------+-------------+
| ``silver`` | ``maroon`` | ``green`` | ``navy``    |
+------------+------------+-----------+-------------+
| ``gray``   | ``yellow`` | ``aqua``  | ``fuchsia`` |
+------------+------------+-----------+-------------+
| ``black``  | ``olive``  | ``teal``  | ``purple``  |
+------------+------------+-----------+-------------+

----

A Python ``dict`` can also be used (Python API only)::

    {
        "duration": 1000,  # ms
        "colors": [
            {"pos": 0, "color": "red"},
            {"pos": 33, "color": "#00FF00"},
            {"pos": 66, "color": (0, 0, 255)},
        ]
    }

.. NOTE::

   A maximum of 14 color stops can be defined in a gradient.


Device Profile
--------------

Example of a rgbgradient value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "logo_color": {
                "label": "Logo LED colors and effects",
                "description": "Set the colors and the effects of the logo LED",
                "cli": ["-c", "--logo-color"],
                "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
                "command": [0x05, 0x00],
                "value_type": "rgbgradient",
                "rgbgradient_header": {
                    "header_length": 28,       # Length of the header excuding command / LED ID
                    "led_id_offsets": [0, 5],  # Offset of the "led_id" fields
                    "duration_offset": 6,      # Offset of the "duration" field
                    "duration_length": 2,      # Length of the "duration" field (in Bytes)
                    "repeat_offset": 22,       # Offset of the "repeat" flag
                    "triggers_offset": 23,     # Offset of the "triggers" field (buttons mask)
                    "color_count_offset": 27,  # Offset of the "color_count" field
                },
                "led_id": 0x01,
                "default": "rgbgradient(duration=1000; colors=0%: #ff0000, 33%: #00ff00, 66%: #0000ff)",
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

   -c LOGO_COLOR, --logo-color LOGO_COLOR
                         Set the colors and the effects of the logo LED (default:
                         rgbgradient(duration=1000; colors=0%: #ff0000, 33%: #00ff00, 66%: #0000ff))

Example of CLI usage::

    rivalcfg --logo-color="rgbgradient(duration=1000; colors=0%: #ff0000, 33%: #00ff00, 66%: #0000ff)"
    rivalcfg --logo-color=red
    rivalcfg --logo-color=FF1800


Functions
---------
"""


import argparse

from ..helpers import uint_to_little_endian_bytearray, merge_bytes
from ..helpers import parse_param_string, REGEXP_PARAM_STRING
from ..color_helpers import is_color, parse_color_string
from ..color_helpers import parse_color_gradient_string


_default_duration = 1000


def _handle_color_tuple(color):
    if len(color) != 3:
        raise ValueError("Not a valid color %s" % str(color))
    for channel in color:
        if type(channel) != int or channel < 0 or channel > 255:
            raise ValueError("Not a valid color %s" % str(color))
    return [
        {
            "pos": 0,
            "color": color,
        }
    ]


def _handle_color_string(color):
    return [
        {
            "pos": 0,
            "color": parse_color_string(color),
        }
    ]


def _handle_rgbgradient_dict(colors):
    duration = _default_duration
    gradient = []

    if "duration" in colors:
        duration = colors["duration"]

    if "colors" in colors:
        for stop in colors["colors"]:
            color = stop["color"]
            if type(color) is str:
                color = parse_color_string(color)
            if type(color) not in [tuple, list] or len(color) != 3:
                raise ValueError("Not a valid color %s" % str(color))
            for channel in color:
                if type(channel) != int or channel < 0 or channel > 255:
                    raise ValueError("Not a valid color %s" % str(color))
            gradient.append(
                {
                    "pos": stop["pos"] if "pos" in stop else 0,
                    "color": color,
                }
            )

    # Smooth gradient (if possible) by adding a final color
    if len(gradient) < 14 and gradient[-1]["pos"] != 100:
        gradient.append(
            {
                "pos": 100,
                "color": gradient[0]["color"],
            }
        )

    return duration, gradient


def _handle_rgbgradient_string(colors):
    gradient_dict = parse_param_string(
        colors,
        value_parsers={
            "rgbgradient": {
                "duration": int,
                "colors": parse_color_gradient_string,
            }
        },
    )

    return _handle_rgbgradient_dict(gradient_dict["rgbgradient"])


def process_value(setting_info, colors):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "rgbgradient" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param str,tuple,list,dict colors: The color(s).
    :rtype: [int]
    """
    header_length = setting_info["rgbgradient_header"]["header_length"]
    led_id_offsets = setting_info["rgbgradient_header"]["led_id_offsets"]
    duration_offset = setting_info["rgbgradient_header"]["duration_offset"]
    duration_length = setting_info["rgbgradient_header"]["duration_length"]
    repeat_offset = setting_info["rgbgradient_header"]["repeat_offset"]
    triggers_offset = setting_info["rgbgradient_header"]["triggers_offset"]
    color_count_offset = setting_info["rgbgradient_header"]["color_count_offset"]

    is_gradient = False
    duration = _default_duration
    repeat = 0x00
    triggers = 0x00
    led_id = setting_info["led_id"]
    gradient = []

    # Color tuple
    if type(colors) in (tuple, list):
        is_gradient = False
        gradient = _handle_color_tuple(colors)

    # Simple color string
    elif type(colors) is str and is_color(colors):
        is_gradient = False
        gradient = _handle_color_string(colors)

    # Color gradient as dict
    elif type(colors) is dict:
        is_gradient = True
        duration, gradient = _handle_rgbgradient_dict(colors)

    # Color gradient as string
    elif is_rgbgradient(colors)[0]:
        is_gradient = True
        duration, gradient = _handle_rgbgradient_string(colors)
    else:
        raise ValueError("Not a valid color or rgbgradient %s" % str(colors))

    # -- handle repeat flag

    if not is_gradient or triggers != 0x00:
        repeat = 0x01

    # -- Check

    if len(gradient) == 0:
        raise ValueError("no color: %s" % str(colors))

    if len(gradient) > 14:
        raise ValueError("a maximum of 14 color stops are allowed")

    # TODO check pos orders

    # -- Generate header

    header = [0x00] * header_length
    header[repeat_offset] = repeat
    header[triggers_offset] = triggers
    header[color_count_offset] = len(gradient)

    for led_id_offset in led_id_offsets:
        header[led_id_offset] = led_id

    offset = 0
    for b in uint_to_little_endian_bytearray(duration, duration_length):
        header[duration_offset + offset] = b
        offset += 1

    # -- Generate body

    body = list(gradient[0]["color"])

    last_real_pos = 0
    for pos, color in [(item["pos"], item["color"]) for item in gradient]:
        real_pos = int(pos * 255 / 100)
        body = merge_bytes(body, color, real_pos - last_real_pos)
        last_real_pos = real_pos

    # --

    return merge_bytes(header, body)


def is_rgbgradient(string):
    """Check if the rgbradient expression is valid.

    :param str string: The string to validate.
    :rtype: (bool, str)

    >>> is_rgbgradient("foo(colors=0:red)")
    (False, "... It must looks like 'rgbgradient(<PARAMS>)'...")
    >>> is_rgbgradient("rgbgradient(duration=1000)")
    (False, "... You must provide at least one color: ...")
    >>> is_rgbgradient("rgbgradient(colors=red)")
    (False, "invalid color gradient...")
    >>> is_rgbgradient("rgbgradient(colors=0:red; foo=bar)")
    (False, "unknown parameter 'foo'...")
    """
    try:
        rgbgradient_dict = parse_param_string(
            string,
            value_parsers={
                "rgbgradient": {
                    "duration": int,
                    "colors": parse_color_gradient_string,
                }
            },
        )
    except ValueError as e:
        return False, str(e)

    if "rgbgradient" not in rgbgradient_dict:
        reason = ""
        reason += "invalid rgbgradient expression. "
        reason += "It must looks like 'rgbgradient(<PARAMS>)'."
        return False, reason

    if (
        "colors" not in rgbgradient_dict["rgbgradient"]
        or len(rgbgradient_dict["rgbgradient"]) == 0
    ):
        reason = ""
        reason += "invalid rgbgradient expression. "
        reason += "You must provide at least one color: "
        reason += "'rgbgradient(colors=<POS>: <COLOR>)'."
        return False, reason

    _allowed_params = ["colors", "duration"]
    for key in rgbgradient_dict["rgbgradient"].keys():
        if key not in _allowed_params:
            reason = ""
            reason += "unknown parameter '%s'. " % key
            reason += "Allowed parameters are %s." % ", ".join(
                ["'%s'" % p for p in _allowed_params]
            )
            return False, reason

    return True, ""


class CheckGradientAction(argparse.Action):
    """Validate colors gradient from CLI"""

    def __call__(self, parser, namespace, value, option_string=None):
        if is_color(value):
            setattr(namespace, self.dest.upper(), value)
            return

        if REGEXP_PARAM_STRING.match(value):
            is_valid, reason = is_rgbgradient(value)

            if is_valid:
                setattr(namespace, self.dest.upper(), value)
                return

            raise argparse.ArgumentError(self, "%s" % reason)

        else:
            raise argparse.ArgumentError(
                self, "not a valid color or rgbgradient: '%s'" % value
            )


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "rgbgradient" type setting to the given CLI arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = "%s (default: %s)" % (
        setting_info["description"],
        str(setting_info["default"]).replace("%", "%%"),
    )
    cli_parser.add_argument(
        *setting_info["cli"],
        dest=setting_name,
        help=description,
        type=str,
        action=CheckGradientAction,
        metavar=setting_name.upper(),
    )



"""
This modules handles udev-related stuff on Linux. It contains function to
generate, check and update rules files.

.. NOTE::

   The functions of this module must only be used with udev-based Linux distro.
"""

import re
import os.path
import subprocess

from .version import VERSION
from .devices import PROFILES


#: Path to the udev rules file
RULES_FILE_PATH = "/etc/udev/rules.d/99-steelseries-rival.rules"


def generate_rules():
    """Generates the content of the udev rules file.

    :rtype: str
    """
    rules = "# Generated by rivalcfg v%s\n" % VERSION
    rules += (
        "# Do not edit this file. It can be regenerated with the following command:\n"
    )
    rules += "# \n"
    rules += "#     rivalcfg --update-udev\n\n"

    for profile in PROFILES.values():
        rules += "# %s\n" % profile["name"]
        rules += (
            'SUBSYSTEM=="hidraw", ATTRS{idVendor}=="%04x", ATTRS{idProduct}=="%04x", MODE="0666"\n'
            % (profile["vendor_id"], profile["product_id"])
        )
        rules += (
            'SUBSYSTEM=="usb", ATTRS{idVendor}=="%04x", ATTRS{idProduct}=="%04x", MODE="0666"\n\n'
            % (profile["vendor_id"], profile["product_id"])
        )

    return rules


def write_rules_file(path=RULES_FILE_PATH):
    """Generates and write the udev rules file at the given place.

    :param str path: The path of the output file.

    :raise PermissionError: The user has not sufficient permissions to write
                            the file.
    """
    rules = generate_rules()
    with open(path, "w") as rules_file:
        rules_file.write(rules)


def reload_rules():
    """Reload udev rules."""
    subprocess.check_output(["udevadm", "control", "--reload-rules"])


def trigger():
    """Trigger udev to take into account the new rules."""
    subprocess.check_output(["udevadm", "trigger"])


def are_rules_up_to_date(rules, current_version=VERSION):
    """Check if the given udev rules are up to date.

    :param str rules: The content of an udev rule file to check.
    :param str current_version: The current rivalcfg version.

    :rtype: bool
    """
    version_regexp = re.compile(r".*rivalcfg\s+v([0-9]+\.[0-9]+\.[0-9]+(.+)?)\s*.*")
    rules_version = None
    if version_regexp.match(rules):
        rules_version = version_regexp.match(rules).group(1)
    return rules_version == current_version


def is_rules_file_up_to_date(path=RULES_FILE_PATH):
    """Check if the given udev rules file is up to date.

    :param str path: The path of the udev rules file.

    :rtype: bool
    """
    if not os.path.isfile(path):
        return False
    with open(path, "r") as rules_file:
        return are_rules_up_to_date(rules_file.read())



"""This module provides access to the profile of each supported device.

Profile spec
------------

A profile is just a ``dict`` that contains information about one or more mouse.
There is one profile file in ``rivalcfg/devices/`` by mouse familly (the "Rival
100" and the "Rival 100 Dota 2 Edition" are mouse of the same familly).

Here is what a profile file looks like::

    from .. import usbhid

    profile = {

        # The name of the mouse familly
        "name": "SteelSeries Rival 100",

        # List of all devices of the familly
        "models": [{
            # The name of a specific device of the familly
            "name": "SteelSeries Rival 100",
            # The vendor ID of the USB device (should always be 0x1038)
            "vendor_id": 0x1038,
            # The product ID of the USB device
            "product_id": 0x1702,
            # The control endpoint of the device (probably 0)
            "endpoint": 0,
        }, {
            "name": "SteelSeries Rival 100 (Dell China)",
            "vendor_id": 0x1038,
            "product_id": 0x170a,
            "endpoint": 0,
            # A device can override default values defined in settings
            "override_defaults": {
                "color": "#FF0000",
            },
        }],

        # All available settings for the mice
        "settings": {

            # A setting of the mouse. The key name is important, it will
            # be used to generates the Mouse class API and in the CLI
            "polling_rate": {
                # A label that will be used one day to build a GUI
                "label": "Polling rate",
                # A short description of the option. Used to generate the CLI
                # help (rivalcfg --help)
                "description": "Set polling rate (Hz)",
                # The name of the CLI options to change this setting.
                "cli": ["-p", "--polling-rate"],
                # The repport type (rivalcfg.usbhid.HID_REPORT_TYPE_OUTPUT most
                # of the time)
                "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
                # The command that should be sent to the USB device to update
                # this setting.
                "command": [0x04, 0x00],
                # A fixed amount of data to send to the mouse (optional).
                # This is required by some mice like the Rival 110.
                "packet_length": 32,
                # If greater that 0, the number of bytes to read from the
                # device after the command. This is required by some wireless
                # mice like the Aerox 3 Wireless in wireless mode.
                "readback_length": 0,
                # The type of value supported by this setting.
                # See the `rivalcfg.handlers` documentation for more
                # information.
                "value_type": "choice",
                # This parameter is specific to the `choice` handler.
                # Look at its documentation for more information.
                "choices": {
                    125: 0x04,
                    250: 0x03,
                    500: 0x02,
                    1000: 0x01,
                },
                # The factory default value of this setting.
                "default": 1000,
            },

            "color": {
                "label": "LED color",
                "description": "Set the mouse backlight color",
                "cli": ["-c", "--color"],
                "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
                "command": [0x05, 0x00],
                "value_type": "rgbcolor",
                "default": "#FF1800"
            },

            # ...

        },

        # The command that allows to persist data in the internal memory
        # of the device
        "save_command": {
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x09, 0x00],
        },

        # (optional) The command to get the firmware version from the device
        "firmware_version": {
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x90, 0x00],
            # The length of the response (bytes)
            "response_length": 2,
        },

    }

At runtime, a separated profile will be generated for each device of a mouse
familly. The profiles will then look like this::

    {

        "name": "SteelSeries Rival 100 (Dell China)",
        "vendor_id": 0x1038,
        "product_id": 0x170a,
        "endpoint": 0,

        "settings": {
            # ...
        },

        "save_command": {
            #...
        },

    }


Module API
----------

.. data:: rivalcfg.devices.PROFILES

    Profiles of each supported devices::

       {
           (vendor_id, product_id): {
               # profile data ...
           },
       }
"""


import os
import types

from . import aerox3  # noqa: F401
from . import aerox3_wireless_wired  # noqa: F401
from . import aerox3_wireless_wireless  # noqa: F401
from . import aerox5_wireless_wired  # noqa: F401
from . import aerox5_wireless_wireless  # noqa: F401
from . import aerox9_wireless_wired  # noqa: F401
from . import aerox9_wireless_wireless  # noqa: F401
from . import kanav2  # noqa: F401
from . import kinzuv2  # noqa: F401
from . import prime  # noqa: F401
from . import prime_wireless_wired  # noqa: F401
from . import prime_wireless_wireless  # noqa: F401
from . import rival3  # noqa: F401
from . import rival3_wireless  # noqa: F401
from . import rival95  # noqa: F401
from . import rival100  # noqa: F401
from . import rival110  # noqa: F401
from . import rival300  # noqa: F401
from . import rival300s  # noqa: F401
from . import rival310  # noqa: F401
from . import rival500  # noqa: F401
from . import rival600  # noqa: F401
from . import rival650  # noqa: F401
from . import rival700  # noqa: F401
from . import sensei310  # noqa: F401
from . import sensei_raw  # noqa: F401
from . import sensei_ten  # noqa: F401
from .. import usbhid


PROFILES = None


class UnsupportedDevice(Exception):
    """Exception raised when the requested device is not supported by rivalcfg."""


def list_plugged_devices():
    """List all plugged devices that are supported by rivalcfg.

    :rtype: generator

    ::

       [
           {"vendor_id": ..., "product_id": ..., "name": ...},
       ]

    If the ``RIVALCFG_PROFILE=vendor_id:product_id`` environment varialbe is
    defined, only the corresponding profile will be listed.
    """
    if "RIVALCFG_PROFILE" in os.environ:
        debug_vendor_id = int(os.environ["RIVALCFG_PROFILE"].split(":")[0], 16)
        debug_product_id = int(os.environ["RIVALCFG_PROFILE"].split(":")[1], 16)
        profile = PROFILES[(debug_vendor_id, debug_product_id)]
        yield {
            "vendor_id": profile["vendor_id"],
            "product_id": profile["product_id"],
            "name": profile["name"],
        }
    for profile in PROFILES.values():
        if usbhid.is_device_plugged(profile["vendor_id"], profile["product_id"]):
            yield {
                "vendor_id": profile["vendor_id"],
                "product_id": profile["product_id"],
                "name": profile["name"],
            }


def get_profile(vendor_id=0x1038, product_id=None):
    """Get the profile of the requested device.

    :param int vendor_id: The vendor id of the device (optional, by default
                          this is set to the SteelSeries vendor id
                          (``0x1038``)).
    :param int product_id: The product id of one of the supported device (e.g.
                           ``0x1702``).
    :raise UnsupportedDevice: The requested device is not supported by
                              rivalcfg.
    :rtype: dict

    >>> from rivalcfg import devices
    >>> devices.get_profile(vendor_id=0x1038, product_id=0x1702)
    {...'name': 'SteelSeries Rival 100'...}
    """
    if not product_id:
        raise ValueError("You must provide a product_id")
    profile_name = (vendor_id, product_id)
    if profile_name not in PROFILES:
        raise UnsupportedDevice(
            "The requested device is not supported (%x:%x)" % profile_name
        )
    return PROFILES[profile_name]


def _generate_profiles():
    """List devices imported in the current module and generates a separated
    profile for each device variation.

    :rtype: dict
    """
    profiles = {}
    for item in [globals()[name] for name in globals()]:
        if not isinstance(item, types.ModuleType):
            continue
        if not hasattr(item, "profile"):
            continue
        for model in item.profile["models"]:
            profile = item.profile.copy()
            profile_name = (model["vendor_id"], model["product_id"])
            del profile["models"]
            for k, v in model.items():
                if k == "override_defaults":
                    continue
                profile[k] = v
            # TODO override_defaults
            profiles[profile_name] = profile
    return profiles


PROFILES = _generate_profiles()



from . import aerox5_wireless_wired


_WIRELESS_FLAG = 0b01000000
_READBACK_LENGTH = 64


def _patch_command(info):
    # Copy the info dict
    result = dict(info)
    # Copy the command list
    result["command"] = list(result["command"])
    # Patch the command
    result["command"][0] = result["command"][0] | _WIRELESS_FLAG
    # Add readback
    result["readback_length"] = _READBACK_LENGTH
    return result


profile = {
    "name": "SteelSeries Aerox 5 Wireless",
    "models": [
        {
            "name": "SteelSeries Aerox 5 Wireless (2.4 GHz wireless mode)",
            "vendor_id": 0x1038,
            "product_id": 0x1852,
            "endpoint": 3,
        },
        {
            "name": "SteelSeries Aerox 5 Wireless Destiny 2 Edition (2.4 GHz wireless mode)",
            "vendor_id": 0x1038,
            "product_id": 0x185C,
            "endpoint": 3,
        },
        {
            "name": "SteelSeries Aerox 5 Wireless Diablo IV Edition (2.4 GHz wireless mode)",
            "vendor_id": 0x1038,
            "product_id": 0x1860,
            "endpoint": 3,
        },
    ],
    "settings": {
        name: _patch_command(info)
        for name, info in aerox5_wireless_wired.profile["settings"].items()
    },
    "battery_level": _patch_command(aerox5_wireless_wired.profile["battery_level"]),
    "save_command": _patch_command(aerox5_wireless_wired.profile["save_command"]),
}



import os
import sys
import platform
import argparse

from . import cli
from . import udev
from . import devices
from . import get_first_mouse


_EPILOG = """
Please report any bug on Github: https://github.com/flozz/rivalcfg/issues
"""


def _check_linux():
    """Run some checks on Linux."""

    # Only run checks on Linux
    if platform.system() != "Linux":
        return

    # Skip check for root as there will be no issue for this user
    if os.getuid() == 0:
        return

    # Skip check if user is trying to update udev rules:
    if "--update-udev" in sys.argv:
        return

    if not os.path.isfile(udev.RULES_FILE_PATH):
        sys.stderr.write(
            "W: udev rules are not installed. You may not be able to open the device using a regular user.\n"
        )
        sys.stderr.write("   Run 'rivalcfg --update-udev' as root to fix.\n\n")
    elif not udev.is_rules_file_up_to_date():
        sys.stderr.write(
            "W: Installed udev rules were generated by an other rivalcfg version.\n"
        )
        sys.stderr.write("   Run 'rivalcfg --update-udev' as root to update.\n\n")


def _render_battery_level(level=None, is_charging=None):
    result = []

    if is_charging is not None:
        result.append("Charging" if is_charging else "Discharging")

    if level is not None:
        result.append("[%-10s] %i %%" % ("=" * int(level / 10), level))

    if is_charging is None and level is None:
        result.append("Unable to get the battery level")

    return " ".join(result)


def main(args=sys.argv[1:]):
    # Display a message when no argument given
    if len(sys.argv) == 1:
        print("USAGE:\n  rivalcfg --help")
        sys.exit(1)

    # On Linux: check udev rules
    _check_linux()

    # Try to open a mouse
    mouse = None
    if (
        "--print-debug" not in sys.argv
        and "--list" not in sys.argv
        and "--version" not in sys.argv
        and "--update-udev" not in sys.argv
        and "--print-udev" not in sys.argv
    ):
        try:
            mouse = get_first_mouse()
        except IOError as error:
            if "--help" not in sys.argv and "-h" not in sys.argv:
                raise error

    cli_parser = argparse.ArgumentParser(prog="rivalcfg", epilog=_EPILOG)
    cli.add_main_cli(cli_parser)

    if mouse:
        mouse_profile = devices.get_profile(
            vendor_id=mouse.vendor_id,
            product_id=mouse.product_id,
        )
        cli.add_mouse_cli(cli_parser, mouse_profile)

    settings = cli_parser.parse_args(args)

    # Print firmware and exit
    if hasattr(settings, "FIRMWARE_VERSION") and settings.FIRMWARE_VERSION:
        print("%s (firmware v%s)" % (mouse.name, mouse.firmware_version))
        sys.exit(0)

    # Print battery level and exit
    if hasattr(settings, "BATTERY_LEVEL") and settings.BATTERY_LEVEL:
        battery_info = mouse.battery
        print(
            _render_battery_level(
                battery_info["level"],
                battery_info["is_charging"],
            )
        )
        sys.exit(0)

    # Reset
    if mouse and settings.RESET:
        mouse.reset_settings()

    # Apply settings
    if mouse:
        for setting_name, value in [(k.lower(), v) for k, v in vars(settings).items()]:
            if value is None:
                continue
            method_name = "set_%s" % setting_name
            if not hasattr(mouse, method_name):
                continue
            getattr(mouse, method_name)(value)

        # Save settings in the internal device memory
        if settings.SAVE:
            mouse.save()


if __name__ == "__main__":
    main()



"""
The "choice" type alows to pick a value from a list and to match it with a
corresponding value in a dict.


Device Profile
--------------

Example of a choice value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "light_effect": {
                "label": "Light effect",
                "description": "Set the light effect",
                "cli": ["-e", "--light-effect"],
                "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
                "command": [0x07, 0x00],
                "value_type": "choice",
                "choices": {
                    "steady": 0x01,
                    "breath": 0x03,
                    1: 0x01,
                    2: 0x02,
                    3: 0x03,
                    4: 0x04,
                },
                "default": "steady",
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

    -e LIGHT_EFFECT, --light-effect=LIGHT_EFFECT
                        Set the light effect (values: steady, breath, 1, 2, 3,
                        4, default: steady)

Example of CLI usage::

    rivalcfg --light-effet=steady


Functions
---------
"""

from .. import helpers


def choices_to_list(choices):
    """Helper function that transforms choices dict to an ordered string list.
    Numeric values are sorted and placed after string values.

    :param dict choices: The dict containing available choices.
    :rtype: list[str]

    >>> choices_to_list({0: 0, 1: 1, 2: 2, "foo": 2})
    ['foo', '0', '1', '2']
    """
    return list(
        map(
            str,
            sorted(choices.keys(), key=lambda v: v if type(v) == int else -1),
        )
    )


def choices_to_string(choices):
    """Helper function that transforms choices dict to a printable string.

    :param dict choices: The dict containing available choices.
    :rtype: str

    >>> choices_to_string({0: 0, 1: 1, 2: 2, "foo": 2})
    'foo, 0, 1, 2'
    """
    return ", ".join(choices_to_list(choices))


def process_value(setting_info, choice):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "choice" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param choice: The selected choice.
    :rtype: list[int]
    """
    choices = {str(k): v for k, v in setting_info["choices"].items()}
    choice = str(choice)
    if choice not in choices:
        raise ValueError(
            "value must be one of [%s]" % choices_to_string(setting_info["choices"])
        )
    value = choices[choice]
    return helpers.merge_bytes(value)


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "choice" type setting to the given CLI arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = "%s (values: %s, default: %s)" % (
        setting_info["description"],
        choices_to_string(setting_info["choices"]),
        str(setting_info["default"]),
    )
    cli_parser.add_argument(
        *setting_info["cli"],
        help=description,
        dest=setting_name.upper(),
        choices=choices_to_list(setting_info["choices"]),
        metavar=setting_name.upper(),
    )



from .. import usbhid


profile = {
    "name": "SteelSeries Rival 650 Wireless",
    "models": [
        {
            "name": "SteelSeries Rival 650 Wireless (wired mode)",
            "vendor_id": 0x1038,
            "product_id": 0x172B,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Rival 650 Wireless (2.4 GHz wireless mode)",
            "vendor_id": 0x1038,
            "product_id": 0x1726,
            "endpoint": 0,
        },
    ],
    "settings": {
        "sensitivity1": {
            "label": "Sensitivity preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x15, 0x01],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 800,
        },
        "sensitivity2": {
            "label": "Sensitivity preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x15, 0x02],
            "value_type": "range",
            "input_range": [100, 12000, 100],
            "output_range": [0x00, 0x77, 1],
            "default": 1600,
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set the polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x17],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },
        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x19],
            "value_type": "buttons",
            # fmt: off
            "buttons": {
                "Button1": {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2": {"id": 0x02, "offset": 0x05, "default": "button2"},
                "Button3": {"id": 0x03, "offset": 0x0A, "default": "button3"},
                "Button4": {"id": 0x04, "offset": 0x0F, "default": "button4"},
                "Button5": {"id": 0x05, "offset": 0x14, "default": "button5"},
                "Button6": {"id": 0x06, "offset": 0x19, "default": "disabled"},
                "Button7": {"id": 0x00, "offset": 0x1E, "default": "dpi"},
            },
            "button_field_length": 5,
            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   0x31,
            "button_scroll_down": 0x32,
            # fmt: on
            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=disabled; button7=dpi)",
        },
        "sleep_timer": {
            "label": "Sleep timer",
            "description": "Set the IDLE time before the mouse goes to sleep mode (minutes)",
            "cli": ["-t", "--sleep-timer"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x2B, 0x01, 0x01, 0x00, 0x00, 0x00],
            "value_type": "range",
            "input_range": [1, 20, 1],
            "output_range": [0x003C, 0x04B0, 60],
            "range_length_byte": 2,
            "default": 5,
        },
    },
    "battery_level": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0xAA, 0x01],
        "response_length": 3,
        "is_charging": lambda data: bool(data[2]),
        "level": lambda data: int(data[0]),
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09],
    },
}



"""
The "range" type alows to pick a value in an input range and transforms it into
a value from an output range. If the input value do not correspond to one of
the input range step, it is rounded to match the nearest step.

For example with an input range like ``[0, 1000, 100]`` and an output range
like ``[0, 10, 1]``, you can have the following pair:

* ``0`` -> ``0``
* ``100`` -> ``1``
* ``110`` -> ``1`` (``110`` rounded to ``100``)
* ``190`` -> ``2`` (``190`` rounded to ``200``)
* ``200`` -> ``2``
* ``300`` -> ``3``
* ...
* ``1000`` -> ``10``


Device Profile
--------------

Example of a range value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "sensitivity1": {
                "label": "Sensitivity preset 1",
                "description": "Set sensitivity preset 1 (DPI)",
                "cli": ["-s", "--sensitivity1"],
                "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
                "command": [0x03, 0x01],
                "value_type": "range",
                "input_range": [200, 7200, 100],
                "output_range": [0x04, 0xA7, 2],
                "range_length_byte": 1,  # optional
                "default": 1000,
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

    -s SENSITIVITY1, --sensitivity1 SENSITIVITY1
                          Set sensitivity preset 1 (DPI) (from 200 to 7200,
                          default: 1000)

Example of CLI usage::

    rivalcfg --sensitivity1=1000


Functions
---------
"""


from ..helpers import uint_to_little_endian_bytearray


def matches_value_in_range(range_start, range_stop, range_step, value):
    """Helper function that matches the value with the nearest value in the
    given range.

    :param int range_start: The start of the range.
    :param int range_stop: The end of the range.
    :param int range_step: The gap between two value in the range.
    :param int value: The value to process.

    :rtype: int

    >>> matches_value_in_range(0, 100, 10, 40)
    40
    >>> matches_value_in_range(0, 100, 10, 42)
    40
    >>> matches_value_in_range(0, 1000, 100, 51)
    100
    >>> matches_value_in_range(42, 1000, 100, 150)
    142
    >>> matches_value_in_range(500, 1000, 100, 100)
    500
    >>> matches_value_in_range(500, 1000, 100, 4000)
    1000
    """
    if value <= range_start:
        return range_start

    if value >= range_stop:
        return range_stop

    delta = (value - range_start) % range_step
    if not delta:
        return value
    else:
        if delta < range_step / 2:
            return value - delta
        else:
            return value - delta + range_step


def custom_range(start, stop, step):
    """Helper function that generates a range of integers but allowing to have
    a float as step.

    I am not very proud of this but the Rival 110 requires a step of ~2.33...

    :param int start: The start of the range.
    :param int stop: The end of the range.
    :param float step: The gap between two value in the range.

    :rtype: generator(int)

    >>> list(custom_range(4, 168, 2.33))
    [4, 6, 8, 10, 13, 15, ...160, 162, 164, 167]
    """
    i = start
    while i < stop:
        yield int(i)
        i += step


def process_range(setting_info, value):
    """Called by the "range" functions to process 'value' with the specified
    range settings in 'setting_info'.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param value: The input value.
    :rtype: int
    """
    input_range = list(
        range(
            setting_info["input_range"][0],
            setting_info["input_range"][1] + 1,
            setting_info["input_range"][2],
        )
    )
    output_range = list(
        custom_range(
            setting_info["output_range"][0],
            setting_info["output_range"][1] + 1,
            setting_info["output_range"][2],
        )
    )

    if len(input_range) != len(output_range):
        raise ValueError("Input range and output range must have the same length")

    matched_value = matches_value_in_range(
        setting_info["input_range"][0],
        setting_info["input_range"][1],
        setting_info["input_range"][2],
        int(value),
    )
    return output_range[input_range.index(matched_value)]


def process_value(setting_info, value):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "range" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param value: The input value.
    :rtype: list[int]
    """
    range_length = 1
    if "range_length_byte" in setting_info:
        range_length = setting_info["range_length_byte"]
    return uint_to_little_endian_bytearray(
        process_range(setting_info, value), range_length
    )


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "range" type setting to the given CLI arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = "%s (from %i to %i, default: %i)" % (
        setting_info["description"],
        setting_info["input_range"][0],
        setting_info["input_range"][1],
        setting_info["default"],
    )
    cli_parser.add_argument(
        *setting_info["cli"],
        help=description,
        dest=setting_name.upper(),
        type=int,
        metavar=setting_name.upper(),
    )



from .. import usbhid


profile = {
    "name": "SteelSeries Rival 3 Wireless",
    "models": [
        {
            "name": "SteelSeries Rival 3 Wireless (2.4 GHz mode)",
            "vendor_id": 0x1038,
            "product_id": 0x1830,
            "endpoint": 3,
        },
    ],
    "settings": {
        "sensitivity": {
            "label": "Sensibility presets",
            "description": "Set sensitivity preset (DPI)",
            "cli": ["-s", "--sensitivity"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x20],
            "value_type": "multidpi_range",
            "input_range": [100, 18000, 100],
            "output_range": [0x00, 0xD6, 1.2],
            "dpi_length_byte": 2,
            "first_preset": 1,
            "count_mode": "number",
            "max_preset_count": 5,
            "default": "400, 800, 1200, 2400, 3200",
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x17],
            "value_type": "choice",
            "choices": {
                125: 0x03,
                250: 0x02,
                500: 0x01,
                1000: 0x00,
            },
            "default": 1000,
        },
        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x19],
            "value_type": "buttons",
            # fmt: off
            "buttons": {
                "Button1": {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2": {"id": 0x02, "offset": 0x05, "default": "button2"},
                "Button3": {"id": 0x03, "offset": 0x0A, "default": "button3"},
                "Button4": {"id": 0x04, "offset": 0x0F, "default": "button4"},
                "Button5": {"id": 0x05, "offset": 0x14, "default": "button5"},
                "Button6": {"id": 0x06, "offset": 0x19, "default": "dpi"},
            },
            "button_field_length": 5,
            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   0x31,
            "button_scroll_down": 0x32,
            # fmt: on
            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=dpi)",
        },
    },
    "battery_level": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0xAA, 0x01],
        "response_length": 3,
        "is_charging": lambda data: bool(data[2]),
        "level": lambda data: int(data[0]),
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09],
    },
}



import pytest

from rivalcfg import mouse
from rivalcfg import usbhid
from rivalcfg import mouse_settings


FAKE_PROFILE = {
    "name": "Fake Mouse",
    "vendor_id": 0x1038,
    "product_id": 0xBAAD,
    "endpoint": 2,
    "settings": {
        "setting1": {
            "label": "Setting 1",
            "description": "A setting with no value",
            "cli": ["-1", "--setting1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0xAA, 0xBB],
            "value_type": None,
        },
        "setting2": {
            "label": "Setting 2",
            "description": "A setting with a simple choice value",
            "cli": ["-2", "--setting2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0xCC],
            "value_type": "choice",
            "choices": {
                "foo": 0x01,
                "bar": 0x02,
            },
            "default": "bar",
        },
        "setting3": {
            "label": "Setting 3",
            "description": "A setting that uses HID feature report",
            "cli": ["-3", "--setting3"],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "command": [0xC0, 0x10, 0x25],
            "value_type": None,
        },
        "setting4": {
            "label": "Setting 4",
            "description": "A setting that use a fixed packet length",
            "cli": ["-4", "--setting4"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x11, 0x22],
            "packet_length": 10,
            "value_type": None,
        },
        "setting5": {
            "label": "Setting 5",
            "description": "A setting with a command suffix",
            "cli": ["-5", "--setting5"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x11, 0x22],
            "command_suffix": [0x33, 0x44],
            "value_type": "choice",
            "choices": {
                42: 0xAA,
            },
            "default": 42,
        },
        "setting6": {
            "label": "Setting 6",
            "description": "A setting with no value",
            "cli": ["-6", "--setting6"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x20, 0x2E],
            "value_type": "none",
        },
        "setting_readback": {
            "label": "Setting with readback",
            "description": "A setting with no value",
            "cli": ["-7", "--setting7"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x20, 0x2F],
            "value_type": "none",
            "readback_length": 42,
        },
        "setting_no_readback": {
            "label": "Setting without readback",
            "description": "A setting with no value",
            "cli": ["-8", "--setting8"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x20, 0x2F],
            "value_type": "none",
            "readback_length": 0,
        },
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x5A, 0x0E],
    },
}

FAKE_PROFILE2 = {
    "name": "Fake Mouse 2",
    "vendor_id": 0x1038,
    "product_id": 0xBAD2,
    "endpoint": 2,
    "settings": {},
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "packet_length": 4,
        "command": [0x5A, 0x0E],
    },
}


class TestGetMouse(object):
    def test_get_mouse_returns_mouse_instance_with_desired_profil(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_DRY", "1")
        rival100 = mouse.get_mouse(vendor_id=0x1038, product_id=0x1702)
        assert rival100.mouse_profile["vendor_id"] == 0x1038
        assert rival100.mouse_profile["product_id"] == 0x1702

    def test_get_mouse_vendor_id_param_optional(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_DRY", "1")
        rival100 = mouse.get_mouse(product_id=0x1702)
        assert rival100.mouse_profile["vendor_id"] == 0x1038
        assert rival100.mouse_profile["product_id"] == 0x1702

    def test_get_mouse_raise_error_if_no_product_id(self):
        with pytest.raises(ValueError):
            mouse.get_mouse()

    @pytest.mark.skip("Not implemented yet")
    def test_get_mouse_raise_error_if_profile_does_not_exists(self):
        pass  # TODO


class TestMouse(object):
    @pytest.fixture
    def mouse(self, monkeypatch):
        return mouse.Mouse(
            usbhid.FakeDevice(),
            FAKE_PROFILE,
            mouse_settings.FakeMouseSettings(0x1038, 0xBAAD, FAKE_PROFILE),
        )

    @pytest.fixture
    def mouse2(self, monkeypatch):
        return mouse.Mouse(
            usbhid.FakeDevice(),
            FAKE_PROFILE2,
            mouse_settings.FakeMouseSettings(0x1038, 0xBAD2, FAKE_PROFILE2),
        )

    def test_name(self, mouse):
        assert mouse.name == "Fake Mouse"

    def test_product_id(self, mouse):
        assert mouse.product_id == 0xBAAD

    def test_firmware_version_tuple(self, mouse):
        assert mouse.firmware_version_tuple == (0,)

    def test_firmware_version(self, mouse):
        assert mouse.firmware_version == "0"

    def test_save(self, mouse):
        mouse.save()
        mouse._hid_device.bytes.seek(0)
        assert mouse._hid_device.bytes.read() == b"\x02\x00\x5A\x0E"

    def test_save_fixed_packet_length(self, mouse2):
        mouse2.save()
        mouse2._hid_device.bytes.seek(0)
        data = mouse2._hid_device.bytes.read()
        assert len(data) == 6
        assert data == b"\x02\x00\x5A\x0E\x00\x00"

    def test_reset_settings(self, mouse):
        mouse.reset_settings()
        mouse._hid_device.bytes.seek(0)
        hid_report = mouse._hid_device.bytes.read()
        assert b"\x02\x00\x5A\x0E" not in hid_report  # No save command
        assert b"\x02\x00\xAA\xBB" in hid_report  # setting1
        assert b"\x02\x00\xCC\x02" in hid_report  # setting2
        assert b"\x03\x00\xC0\x10\x25" in hid_report  # setting3

    def test_repr(self, mouse):
        assert str(mouse) == "<Mouse Fake Mouse (1038:baad:02)>"
        assert repr(mouse) == "<Mouse Fake Mouse (1038:baad:02)>"

    @pytest.mark.parametrize(
        "report_type,expected_hid_report",
        [
            (usbhid.HID_REPORT_TYPE_OUTPUT, b"\x02\x00"),
            (usbhid.HID_REPORT_TYPE_FEATURE, b"\x03\x00"),
        ],
    )
    def test__hid_write_with_a_valid_report_type(
        self, mouse, report_type, expected_hid_report
    ):
        mouse._hid_write(report_type=report_type)
        mouse._hid_device.bytes.seek(0)
        assert mouse._hid_device.bytes.read() == expected_hid_report

    def test__hid_write_with_a_wrong_report_type(self, mouse):
        with pytest.raises(ValueError):
            mouse._hid_write(report_type=0x42)

    def test__hid_write_with_data(self, mouse):
        mouse._hid_write(
            report_type=usbhid.HID_REPORT_TYPE_OUTPUT,
            report_id=0x42,
            data=[0xDA, 0x7A],
        )
        mouse._hid_device.bytes.seek(0)
        assert mouse._hid_device.bytes.read() == b"\x02\x42\xDA\x7A"

    def test__hid_write_with_fixed_packet_length(self, mouse):
        mouse._hid_write(packet_length=4)
        mouse._hid_device.bytes.seek(0)
        data = mouse._hid_device.bytes.read()
        assert len(data) == 6
        assert data == b"\x02\x00\x00\x00\x00\x00"

    # Virtual methods dependent to the loaded profile

    def test_set_setting1_is_available(self, mouse):
        assert hasattr(mouse, "set_setting1")

    def test_set_setting1(self, mouse):
        mouse.set_setting1()
        mouse._hid_device.bytes.seek(0)
        assert mouse._hid_device.bytes.read() == b"\x02\x00\xAA\xBB"

    def test_set_setting2_is_available(self, mouse):
        assert hasattr(mouse, "set_setting2")

    @pytest.mark.parametrize(
        "value,expected_hid_report",
        [
            ("foo", b"\x02\x00\xCC\x01"),
            ("bar", b"\x02\x00\xCC\x02"),
        ],
    )
    def test_setting2(self, mouse, value, expected_hid_report):
        mouse.set_setting2(value)
        mouse._hid_device.bytes.seek(0)
        assert mouse._hid_device.bytes.read() == expected_hid_report

    def test_set_setting3_is_available(self, mouse):
        assert hasattr(mouse, "set_setting3")

    def test_set_setting3(self, mouse):
        mouse.set_setting3()
        mouse._hid_device.bytes.seek(0)
        assert mouse._hid_device.bytes.read() == b"\x03\x00\xC0\x10\x25"

    def test_set_setting4_packet_length(self, mouse):
        mouse.set_setting4()
        mouse._hid_device.bytes.seek(0)
        data = mouse._hid_device.bytes.read()
        assert len(data) == 12
        assert data == b"\x02\x00\x11\x22\x00\x00\x00\x00\x00\x00\x00\x00"

    def test_set_setting5_command_suffix(self, mouse):
        mouse.set_setting5(42)
        mouse._hid_device.bytes.seek(0)
        data = mouse._hid_device.bytes.read()
        assert data == b"\x02\x00\x11\x22\xAA\x33\x44"

    def test_an_unexisting_setting(self, mouse):
        assert not hasattr(mouse, "set_xxx")
        with pytest.raises(AttributeError):
            mouse.set_xxx()

    def test_readback(self, mouse):
        response = mouse.set_setting_readback()
        assert len(response) == 42

    def test_no_readback(self, mouse):
        response = mouse.set_setting_no_readback()
        assert response is None
        response = mouse.set_setting6()
        assert response is None



import time

from . import usbhid
from . import devices
from . import handlers
from . import helpers
from . import mouse_settings


def get_mouse(vendor_id=0x1038, product_id=None):
    """Get a :class:`Mouse` instance to manipulate requested device.

    :param int vendor_id: The vendor id of the device (optional, by default
                          this is set to the SteelSeries vendor id
                          (``0x1038``)).
    :param int product_id: The product id of one of the supported device (e.g.
                           ``0x1702``).
    :raise rivalcfg.devices.UnsupportedDevice: The requested device is not
                                               supported by rivalcfg.
    :rtype: Mouse

    >>> from rivalcfg.mouse import get_mouse
    >>> get_mouse(vendor_id=0x1038, product_id=0x1702)
    <Mouse SteelSeries Rival 100 (1038:1702:00)>
    """
    if not product_id:
        raise ValueError("You must define the 'product_id' parameter")

    profile = devices.get_profile(vendor_id, product_id)
    settings = mouse_settings.get_mouse_settings(
        vendor_id,
        product_id,
        profile,
    )

    hid_device = usbhid.open_device(vendor_id, product_id, profile["endpoint"])

    return Mouse(hid_device, profile, settings)


class Mouse:

    """Generic class to handle any supported mouse.

    .. NOTE::

       Additional methods are available in this class depending on the loaded
       profile. Read device specific documentation for more information.

    .. WARNING::

       You should not instanciate this class yourself. Use the
       :func:`get_mouse` factory function instead.

    :param hid_device: The HID device to write in (provided by the
                        :func:`rivalcfg.usbhid.open_device`).
    :param mouse_profile: One of the rivalcfg mouse profile (provided by
                            :func:`rivalcfg.devices.get_profile`).

    >>> from rivalcfg import usbhid
    >>> from rivalcfg import devices
    >>> from rivalcfg.mouse import Mouse
    >>> from rivalcfg.mouse_settings import get_mouse_settings
    >>> profile = devices.get_profile(vendor_id=0x1038, product_id=0x1702)
    >>> settings = get_mouse_settings(0x1038, 0x1702, profile)
    >>> Mouse(
    ...     usbhid.open_device(vendor_id=0x1038, product_id=0x1702, endpoint=0),
    ...     profile,
    ...     settings,
    ... )
    <Mouse SteelSeries Rival 100 (1038:1702:00)>
    """

    #: The mouse settings (``rivalcfg.devices.*``)
    mouse_profile = None

    #: The mouse settings (:class:`rivalcfg.mouse_settings.MouseSettings`)
    mouse_settings = None

    def __init__(self, hid_device, mouse_profile, mouse_settings):
        """Constructor."""
        self._hid_device = hid_device
        self.mouse_profile = mouse_profile
        self.mouse_settings = mouse_settings

    @property
    def name(self):
        """The mouse name."""
        return self.mouse_profile["name"]

    @property
    def vendor_id(self):
        """The mouse vendor id."""
        return self.mouse_profile["vendor_id"]

    @property
    def product_id(self):
        """The mouse product id."""
        return self.mouse_profile["product_id"]

    @property
    def firmware_version_tuple(self):
        """The firmware version of the device as a tuple (e.g.``(1, 33)``,
        ``(0,)`` if not available).
        """
        if "firmware_version" not in self.mouse_profile:
            return (0,)
        self._hid_write(
            self.mouse_profile["firmware_version"]["report_type"],
            data=self.mouse_profile["firmware_version"]["command"],
        )
        version = self._hid_device.read(
            self.mouse_profile["firmware_version"]["response_length"],
            timeout_ms=200,
        )
        if not version:
            return (0,)
        return tuple(version)

    @property
    def firmware_version(self):
        """The firmware version as an human readable string (e.g. ``"1.33"``,
        ``"0"`` if not available).
        """
        return ".".join([str(i) for i in self.firmware_version_tuple])

    @property
    def battery(self):
        """Information about the device battery.

        :rtype: dict
        :return: ``{"is_charging": True|False|None, "level": int(0-100)|None}``.

        .. NOTE::

           A value of ``None`` means the featue is not supported.
        """
        result = {
            "is_charging": None,
            "level": None,
        }
        if "battery_level" not in self.mouse_profile:
            return result

        self._hid_write(
            self.mouse_profile["battery_level"]["report_type"],
            data=self.mouse_profile["battery_level"]["command"],
        )
        data = self._hid_device.read(
            self.mouse_profile["battery_level"]["response_length"],
            timeout_ms=200,
        )

        try:
            if "is_charging" in self.mouse_profile["battery_level"]:
                result["is_charging"] = self.mouse_profile["battery_level"][
                    "is_charging"
                ](data)
        except Exception:
            pass

        try:
            if "level" in self.mouse_profile["battery_level"]:
                result["level"] = self.mouse_profile["battery_level"]["level"](data)
        except Exception:
            pass

        return result

    def reset_settings(self):
        """Sets all settings to their factory default values."""
        for name, setting_info in self.mouse_profile["settings"].items():
            method_name = "set_%s" % name
            method = getattr(self, method_name)
            if (
                "value_type" in setting_info
                and setting_info["value_type"]
                and setting_info["value_type"] != "none"
            ):
                method(setting_info["default"])
            else:
                method()

    def save(self):
        """Save current config to the mouse internal memory."""
        # This should never happen... But who knows...
        if (
            "save_command" not in self.mouse_profile
            or not self.mouse_profile["save_command"]
        ):
            raise Exception("This mouse does not provide any save command.")

        packet_length = 0
        if "packet_length" in self.mouse_profile["save_command"]:
            packet_length = self.mouse_profile["save_command"]["packet_length"]

        self._hid_write(
            report_type=self.mouse_profile["save_command"]["report_type"],
            data=self.mouse_profile["save_command"]["command"],
            packet_length=packet_length,
        )

        response = None
        if (
            "readback_length" in self.mouse_profile["save_command"]
            and self.mouse_profile["save_command"]["readback_length"]
        ):
            response = self._hid_device.read(
                self.mouse_profile["save_command"]["readback_length"],
                timeout_ms=200,
            )

        self.mouse_settings.save()

        return response

    def close(self):
        """Close the device.

        .. WARNING::

           Once called, any access of the Mouse class properties or function
           may raise an error.
        """
        self._hid_device.close()

    def _hid_write(
        self,
        report_type=usbhid.HID_REPORT_TYPE_OUTPUT,
        report_id=0x00,
        data=[],
        packet_length=0,
    ):
        """
        Write data to the device.

        :param int report_type: The HID report type
                                (:data:`rivalcfg.usbhid.HID_REPORT_TYPE_OUTPUT`
                                or
                                :data:`rivalcfg.usbhid.HID_REPORT_TYPE_FEATURE`).
        :param int report_id: The id of the report (always ``0x00``).
        :param list(int) data: The data to send to the mouse.
        :param int packet_length: The fixed length of the packet that will be
                                  sent to the device (default: ``0`` (no fixed
                                  length)).

        :raises ValueError: Invalid report type, or HID device not openned.
        """
        if packet_length:
            bytes_ = bytearray(
                helpers.merge_bytes(
                    report_id, data, [0x00] * (packet_length - len(data))
                )
            )
        else:
            bytes_ = bytearray(helpers.merge_bytes(report_id, data))
        if report_type == usbhid.HID_REPORT_TYPE_OUTPUT:
            self._hid_device.write(bytes_)
        elif report_type == usbhid.HID_REPORT_TYPE_FEATURE:
            self._hid_device.send_feature_report(bytes_)
        else:
            raise ValueError("Invalid HID report type: %2x" % report_type)

        # Avoids sending multiple commands to quickly
        time.sleep(0.05)

    def __getattr__(self, name):
        # Handle every set_xxx methods generated from device's profiles

        if not name.startswith("set_"):
            raise AttributeError("Mouse instance has no attribute '%s'" % name)

        setting_name = name[4:]

        if setting_name not in self.mouse_profile["settings"]:
            raise AttributeError("Mouse instance has no attribute '%s'" % name)

        setting_info = self.mouse_profile["settings"][setting_name]

        handler_name = None

        if "value_type" in setting_info and setting_info["value_type"]:
            handler_name = setting_info["value_type"]
            if handler_name not in helpers.module_ls(handlers):
                raise ValueError(
                    "Unknown handler '%s' for '%s' setting of the %s"
                    % (
                        handler_name,
                        setting_name,
                        self.mouse_profile["name"],
                    )
                )

        packet_length = 0
        if "packet_length" in setting_info:
            packet_length = setting_info["packet_length"]

        suffix = []
        if "command_suffix" in setting_info:
            suffix = setting_info["command_suffix"]

        def _exec_command(*args):
            data = []
            if handler_name:
                data = getattr(handlers, handler_name).process_value(
                    setting_info, *args
                )
            # Write data to the device
            self._hid_write(
                report_type=setting_info["report_type"],
                data=helpers.merge_bytes(setting_info["command"], data, suffix),
                packet_length=packet_length,
            )
            # Readback when required
            response = None
            if "readback_length" in setting_info and setting_info["readback_length"]:
                response = self._hid_device.read(
                    setting_info["readback_length"],
                    timeout_ms=200,
                )
            # Save settings
            if len(args) == 1:
                self.mouse_settings.set(setting_name, args[0])
            else:
                self.mouse_settings.set(setting_name, args)
            #
            return response

        return _exec_command

    def __repr__(self):
        return "<Mouse %s (%04x:%04x:%02x)>" % (
            self.mouse_profile["name"],
            self.mouse_profile["vendor_id"],
            self.mouse_profile["product_id"],
            self.mouse_profile["endpoint"],
        )

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        self.close()



from .. import usbhid


_BATTERY_CHARGING_FLAG = 0b10000000


profile = {
    "name": "SteelSeries Aerox 5 Wireless",
    "models": [
        {
            "name": "SteelSeries Aerox 5 Wireless (wired mode)",
            "vendor_id": 0x1038,
            "product_id": 0x1854,
            "endpoint": 3,
        },
        {
            "name": "SteelSeries Aerox 5 Wireless Destiny 2 Edition (wired mode)",
            "vendor_id": 0x1038,
            "product_id": 0x185E,
            "endpoint": 3,
        },
        {
            "name": "SteelSeries Aerox 5 Wireless Diablo IV Edition (wired mode)",
            "vendor_id": 0x1038,
            "product_id": 0x1862,
            "endpoint": 3,
        },
    ],
    "settings": {
        "sensitivity": {
            "label": "Sensibility presets",
            "description": "Set sensitivity preset (DPI)",
            "cli": ["-s", "--sensitivity"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x2D],
            "value_type": "multidpi_range",
            "input_range": [100, 18000, 100],
            "output_range": [0x00, 0xD6, 1.2],
            "dpi_length_byte": 1,
            "first_preset": 0,
            "count_mode": "number",
            "max_preset_count": 5,
            "default": "400, 800, 1200, 2400, 3200",
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x2B],
            "value_type": "choice",
            "choices": {
                125: 0x03,
                250: 0x02,
                500: 0x01,
                1000: 0x00,
            },
            "default": 1000,
        },
        "z1_color": {
            "label": "Strip top LED color",
            "description": "Set the color of the top LED",
            "cli": ["--top-color", "--z1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x21, 0x01, 0x00],
            "value_type": "rgbcolor",
            "default": "red",
        },
        "z2_color": {
            "label": "Strip middle LED color",
            "description": "Set the color of the middle LED",
            "cli": ["--middle-color", "--z2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x21, 0x01, 0x01],
            "value_type": "rgbcolor",
            "default": "lime",
        },
        "z3_color": {
            "label": "Strip bottom LED color",
            "description": "Set the color of the bottom LED",
            "cli": ["--bottom-color", "--z3"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x21, 0x01, 0x02],
            "value_type": "rgbcolor",
            "default": "blue",
        },
        "reactive_color": {
            "label": "Reactive color",
            "description": "Set the color of the LEDs in reaction to a button click",
            "cli": ["-a", "--reactive-color"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x26],
            "value_type": "reactive_rgbcolor",
            "default": "off",
        },
        "sleep_timer": {
            "label": "Sleep timer",
            "description": "Set the IDLE time before the mouse goes to sleep mode (minutes, 0 = disable)",
            "cli": ["-t", "--sleep-timer"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x29],
            "value_type": "range",
            "input_range": [0, 20, 1],
            "output_range": [0x000000, 0x124F80, 60000],
            "range_length_byte": 3,
            "default": 5,
        },
        "dim_timer": {
            "label": "Dim timer",
            "description": "Set the IDLE time before the mouse light is dimmed (seconds, 0 = disable)",
            "cli": ["-T", "--dim-timer"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x23, 0x0F, 0x01, 0x00, 0x00],
            "value_type": "range",
            "input_range": [0, 1200, 1],
            "output_range": [0x000000, 0x124F80, 1000],
            "range_length_byte": 3,
            "default": 30,
        },
        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttonsX",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x2A],
            "value_type": "buttons",
            # fmt: off
            "buttons": {
                "Button1":    {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2":    {"id": 0x02, "offset": 0x05, "default": "button2"},
                "Button3":    {"id": 0x03, "offset": 0x0A, "default": "button3"},
                "Button4":    {"id": 0x04, "offset": 0x0F, "default": "button4"},
                "Button5":    {"id": 0x05, "offset": 0x14, "default": "button5"},
                "Button6":    {"id": 0x06, "offset": 0x19, "default": "dpi"},
                "Button7":    {"id": 0x00, "offset": 0x1E, "default": "disabled"},
                "Button8":    {"id": 0x00, "offset": 0x23, "default": "disabled"},
                "Button9":    {"id": 0x00, "offset": 0x28, "default": "disabled"},
                "ScrollUp"  : {"id": 0x31, "offset": 0x2D, "default": "scrollup"},
                "ScrollDown": {"id": 0x32, "offset": 0x32, "default": "scrolldown"},
            },
            "button_field_length": 5,
            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   None,
            "button_scroll_down": None,
            # fmt: on
            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=dpi; button7=disabled; button8=disabled; button9=disabled; scrollup=scrollup; scrolldown=scrolldown; layout=qwerty)",
        },
        "rainbow_effect": {
            "label": "rainbow effect",
            "description": "Enable the rainbow effect (can be disabled by setting a color)",
            "cli": ["-e", "--rainbow-effect"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x22, 0xFF],
            "value_type": "none",
        },
        "default_lighting": {
            "label": "Default lighting",
            "description": "Set default lighting at mouse startup",
            "cli": ["-d", "--default-lighting"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x27],
            "value_type": "choice",
            "choices": {
                "off": [0x00, 0x00],
                "reactive": [0x00, 0x01],
                "rainbow": [0x01, 0x00],
                "reactive-rainbow": [0x01, 0x01],
            },
            "default": "rainbow",
        },
    },
    "battery_level": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x92],
        "response_length": 2,
        "is_charging": lambda data: bool(data[1] & _BATTERY_CHARGING_FLAG),
        "level": lambda data: ((data[1] & ~_BATTERY_CHARGING_FLAG) - 1) * 5,
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x11, 0x00],
    },
}



import json

import pytest

from rivalcfg import mouse_settings


FAKE_PROFILE = {
    "settings": {
        "setting1": {
            "default": "foo",
        },
        "setting2": {
            "default": "bar",
        },
        "setting3": {
            "value_type": "none",
        },
    },
}


class Test_get_xdg_config_home(object):
    def test_environ_not_set(self, monkeypatch):
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
        monkeypatch.setenv("HOME", "/tmp")
        assert mouse_settings.get_xdg_config_home() == "/tmp/.config"

    def test_environ_set(self, monkeypatch):
        monkeypatch.setenv("XDG_CONFIG_HOME", "/tmp/foo/bar")
        assert mouse_settings.get_xdg_config_home() == "/tmp/foo/bar"


class Test_get_settings_path(object):
    def test_settings_file_path(self, monkeypatch):
        monkeypatch.setenv("XDG_CONFIG_HOME", "/tmp/foo/bar")
        assert (
            mouse_settings.get_settings_path(0x1038, 0xBAAD)
            == "/tmp/foo/bar/rivalcfg/1038_baad.device.json"
        )


class Test_FakeMouseSettings(object):
    @pytest.fixture
    def mouse_settings(self):
        return mouse_settings.FakeMouseSettings(0x1038, 0xBAAD, FAKE_PROFILE)

    def test_get_default_values(self, mouse_settings):
        assert mouse_settings.get_default_values() == {
            "setting1": "foo",
            "setting2": "bar",
        }

    def test_get_set_with_existing_setting(self, mouse_settings):
        assert mouse_settings.get("setting1") == "foo"
        mouse_settings.set("setting1", "test")
        assert mouse_settings.get("setting1") == "test"

    def test_get_with_non_existing_setting(self, mouse_settings):
        with pytest.raises(KeyError):
            mouse_settings.get("foobar")

    def test_set_with_non_existing_setting(self, mouse_settings):
        with pytest.raises(KeyError):
            mouse_settings.set("foobar", "test")


class Test_MouseSettings(object):
    def test_save_load_defaults(self, monkeypatch, tmpdir):
        monkeypatch.delenv("DEBUG_DRY", raising=False)
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmpdir))
        ms = mouse_settings.MouseSettings(0x1038, 0xBAAD, FAKE_PROFILE)
        ms.save()
        mouse_settings.MouseSettings(0x1038, 0xBAAD, FAKE_PROFILE)

    def test_save_load(self, monkeypatch, tmpdir):
        monkeypatch.delenv("DEBUG_DRY", raising=False)
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmpdir))
        ms = mouse_settings.MouseSettings(0x1038, 0xBAAD, FAKE_PROFILE)
        ms.set("setting1", "test1")
        ms.save()
        ms2 = mouse_settings.MouseSettings(0x1038, 0xBAAD, FAKE_PROFILE)
        assert ms2.get("setting1") == "test1"

    def test_none_value_type_settings_return_none(self, monkeypatch, tmpdir):
        monkeypatch.delenv("DEBUG_DRY", raising=False)
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmpdir))
        ms = mouse_settings.MouseSettings(0x1038, 0xBAAD, FAKE_PROFILE)
        ms.set("setting3", "foo")
        assert ms.get("setting3") is None

    def test_none_value_type_settings_not_saved(self, monkeypatch, tmpdir):
        monkeypatch.delenv("DEBUG_DRY", raising=False)
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmpdir))
        ms = mouse_settings.MouseSettings(0x1038, 0xBAAD, FAKE_PROFILE)
        ms.set("setting3", "foo")
        ms.save()
        with open(mouse_settings.get_settings_path(0x1038, 0xBAAD), "r") as file_:
            saved_data = json.load(file_)
        assert "setting3" not in saved_data["default"]



"""
The "none" type is used for command with no arguments.


Device Profile
--------------

Example of a "none" value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "rainbow_effect": {
                "label": "rainbow effect",
                "description": "Enable the rainbow effect (can be disabled by setting a color)",
                "cli": ["-e", "--rainbow-effect"],
                "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
                "command": [0x22, 0xFF],
                "value_type": "none",
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

    -e, --rainbow-effect  Enable the rainbow effect (can be disabled by setting
                          a color)

Example of CLI usage::

    rivalcfg --rainbow-effect


Functions
---------
"""


def process_value(setting_info, *args):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "none" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :rtype: list[]
    """
    return []


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "none" type setting to the given CLI arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    cli_parser.add_argument(
        *setting_info["cli"],
        help=setting_info["description"],
        action="store_const",
        const=True,
    )



from .. import usbhid


profile = {
    "name": "SteelSeries Sensei [RAW]",
    "models": [
        {
            "name": "SteelSeries Sensei [RAW]",
            "vendor_id": 0x1038,
            "product_id": 0x1369,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Sensei [RAW] Diablo III Edition",
            "vendor_id": 0x1038,
            "product_id": 0x1362,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Sensei [RAW] Guild Wars 2 Edition",
            "vendor_id": 0x1038,
            "product_id": 0x136D,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Sensei [RAW] CoD Black Ops II Edition",
            "vendor_id": 0x1038,
            "product_id": 0x136F,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Sensei [RAW] World of Tanks Edition",
            "vendor_id": 0x1038,
            "product_id": 0x1380,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Sensei [RAW] Heroes of the Storm Edition",
            "vendor_id": 0x1038,
            "product_id": 0x1390,
            "endpoint": 0,
        },
    ],
    "settings": {
        "sensitivity1": {
            "label": "Sensibility preset 1",
            "description": "Set sensitivity preset 1 (DPI)",
            "cli": ["-s", "--sensitivity1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x01],
            "value_type": "range",
            "input_range": [90, 5670, 90],
            "output_range": [0x01, 0x3F, 1],
            "default": 1620,
        },
        "sensitivity2": {
            "label": "Sensitivity preset 2",
            "description": "Set sensitivity preset 2 (DPI)",
            "cli": ["-S", "--sensitivity2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x03, 0x02],
            "value_type": "range",
            "input_range": [90, 5670, 90],
            "output_range": [0x01, 0x3F, 1],
            "default": 3240,
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },
        "led_brightness": {
            "label": "LED Brightness",
            "description": "Set the brightness of the LEDs",
            "cli": ["-l", "--led-brightness"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x05, 0x01],
            "value_type": "choice",
            "choices": {
                "off": 0x01,
                "low": 0x02,
                "medium": 0x03,
                "high": 0x04,
            },
            "default": "off",
        },
        "light_effect": {
            "label": "Light effect",
            "description": "Set the light effect",
            "cli": ["-e", "--light-effect"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x07, 0x01],
            "value_type": "choice",
            "choices": {
                "steady": 0x01,
                "breath": 0x03,
                1: 0x01,
                2: 0x02,
                3: 0x03,
                4: 0x04,
                "trigger": 0x05,
            },
            "default": "breath",
        },
        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x31, 0x00],
            "value_type": "buttons",
            "buttons": {
                "Button1": {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2": {"id": 0x02, "offset": 0x03, "default": "button2"},
                "Button3": {"id": 0x03, "offset": 0x06, "default": "button3"},
                "Button4": {"id": 0x04, "offset": 0x09, "default": "button4"},
                "Button5": {"id": 0x05, "offset": 0x0C, "default": "button5"},
                "Button6": {"id": 0x06, "offset": 0x12, "default": "PageDown"},
                "Button7": {"id": 0x07, "offset": 0x0F, "default": "PageUp"},
                "Button8": {"id": 0x08, "offset": 0x15, "default": "dpi"},
            },
            # fmt: off
            "button_disable":     0x00,
            "button_keyboard":    0x10,
            "button_multimedia":  0x11,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   None,
            "button_scroll_down": None,
            # fmt: on
            "button_field_length": 3,
            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=PageDown; button7=PageUp; button8=dpi; layout=qwerty)",
        },
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },
}



"""
The "buttons" type handles buttons mapping on device that supports it.

Buttons syntax example::

    default
    buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=dpi; layout=qwerty)
    buttons(Button6=DPI)  # all other buttons will be reset to their default values


Buttons values
--------------

Buttons can be mapped to different kind of values:

* ``default`` (to explicitly set the button to its default value)
* Special actions,
* Mouse buttons,
* Multimedia keys,
* Keyboard keys.


Special actions
~~~~~~~~~~~~~~~

* ``disabled``: disable this button,
* ``dpi``: use this button to switch between DPI presets,
* ``ScrollUp`` (not available on all devices)
* ``ScrollDown`` (not available on all devices)

Example::

    buttons(button6=dpi; button5=disabled)


Mouse buttons
~~~~~~~~~~~~~

Mouse buttons can be remapped using an other mouse button.

For example, this swap the button 1 and 2 of the mouse::

    buttons(button1=button2; button2=button1)

.. WARNING::

    Be careful to map the ``button1`` somewhere, else you will not be able to
    click! ;)


Multimedia keys
~~~~~~~~~~~~~~~

Example::

    buttons(button8=PlayPause)

List of the available multimedia keys:

* https://github.com/flozz/rivalcfg/blob/master/rivalcfg/handlers/buttons/layout_multimedia.py


Keyboard keys
~~~~~~~~~~~~~

Keyboard keys depends on the selected layout.

Examples::

    buttons(layout=QWERTY; button2=A)

.. NOTE::

    Due to some syntax limitation, some keys cannot be mapped used their own
    symbols. For example, if you want to bind ``;`` to the third button, you
    will have to use an alias::

        buttons(layout=QWERTY; button3=semicolon)

    More examples::

        buttons(layout=QWERTY; button3=semi; button4=equal)


Layouts
-------

The layout used for mapping keyboard keys can be specified using the ``layout``
parameter::

    buttons(layout=QWERTY)

Currently, the following layouts are supported:

* ``QWERTY`` (`see available keys <https://github.com/flozz/rivalcfg/blob/master/rivalcfg/handlers/buttons/layout_qwerty.py>`_)

If not specified, the ``QWERTY`` layout is used.


Device Profile
--------------

Example of a buttons value type in a device profile:

::

    profile = {

        # ...

        "settings": {

        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x31, 0x00],
            "value_type": "buttons",

            "buttons": {
                "Button1": {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2": {"id": 0x02, "offset": 0x05, "default": "button2"},
                "Button3": {"id": 0x03, "offset": 0x0A, "default": "button3"},
                "Button4": {"id": 0x04, "offset": 0x0F, "default": "button4"},
                "Button5": {"id": 0x05, "offset": 0x14, "default": "button5"},
                "Button6": {"id": 0x06, "offset": 0x19, "default": "dpi"},
            },

            "button_field_length": 5,

            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   0x31,
            "button_scroll_down": 0x32,

            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=dpi; layout=qwerty)",
        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

  -b BUTTONS_MAPPING, --buttons BUTTONS_MAPPING
                        Set the mapping of the buttons (default: buttons(button1=button1;
                        button2=button2; button3=button3; button4=button4; button5=button5;
                        button6=dpi; layout=qwerty))

Example of CLI usage::

    rivalcfg --buttons="buttons(layout=qwerty; button1=button1; button6=DPI)"


Functions
---------
"""


import argparse

from ...helpers import parse_param_string, REGEXP_PARAM_STRING
from . import layout_multimedia
from . import layout_qwerty


#: Available layouts
LAYOUTS = {
    "qwerty": layout_qwerty,
    # TODO azerty
}


def build_layout(layout):
    """Generates an usable layout from a layout module containing key mapping
    and aliases.

    :param module layout: The layour (e.g.
                          ``rivalcfg.handler.buttons.layout_qwerty``)
    :rtype dict:
    """
    full_layout = {k.lower(): v for k, v in layout.layout.items()}

    for alias, ref in layout.aliases.items():
        if ref not in layout.layout:
            raise ValueError(
                "Wrong alias: '%s' aliases '%s' but '%s' is not in the layout"
                % (alias, ref, ref)
            )
        full_layout[alias.lower()] = layout.layout[ref]

    return full_layout


def is_buttons(string, setting_info):
    """Checks if the regbradient expression is valid.

    :param str string: The string to validate.
    :param dict setting_info: the settings info from the mouse profile.
    :rtype: (bool, str)
    """
    available_buttons = {k.lower(): v for k, v in setting_info["buttons"].items()}

    try:
        buttons_dict = parse_param_string(string)
    except ValueError as e:
        return False, str(e)

    if "buttons" not in buttons_dict:
        return (
            False,
            "Buttons expression must looks like buttons(<BUTTON>=<VALUE>; <BUTTON_N>=<VALUE_N>)",
        )

    for key, value in [
        (k.lower(), v.lower()) for k, v in buttons_dict["buttons"].items()
    ]:
        if key == "layout":
            if value not in LAYOUTS:
                return False, "Unknown layout '%s'" % value
        else:
            if key not in available_buttons:
                return False, "This mouse have no button named '%s'" % key

    return True, ""


def process_value(setting_info, mapping):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "buttons" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param str,dict mapping: The mapping of the mouse buttons.
    :rtype: [int]
    """
    # -- Parse input values

    if type(mapping) is str and REGEXP_PARAM_STRING.match(mapping):
        is_valid, reason = is_buttons(mapping, setting_info)
        if not is_valid:
            raise ValueError(reason)
        mapping = parse_param_string(mapping)
    elif type(mapping) is str and mapping.lower() == "default":
        mapping = {"buttons": {}}
    elif type(mapping) is dict:
        pass
    else:
        raise ValueError("Invalid input value '%s'" % str(mapping))

    # -- Initialize mouse buttons

    buttons = {k.lower(): dict(v) for k, v in setting_info["buttons"].items()}

    # -- Initialise special values

    button_keyboard = setting_info["button_keyboard"]
    button_multimedia = setting_info["button_multimedia"]

    button_special = {}

    if setting_info["button_disable"] is not None:
        button_special["disable"] = setting_info["button_disable"]
        button_special["disabled"] = setting_info["button_disable"]

    if setting_info["button_dpi_switch"] is not None:
        button_special["dpi"] = setting_info["button_dpi_switch"]

    if setting_info["button_scroll_up"] is not None:
        button_special["scrollup"] = setting_info["button_scroll_up"]

    if setting_info["button_scroll_down"] is not None:
        button_special["scrolldown"] = setting_info["button_scroll_down"]
        button_special["scrolldwn"] = setting_info["button_scroll_down"]
        button_special["scrolldn"] = setting_info["button_scroll_down"]

    # -- Initialize keyboard layout

    keyboard_layout = {}
    keyboard_layout_name = "qwerty"

    if button_keyboard is not None:
        if "layout" in mapping["buttons"]:
            keyboard_layout_name = mapping["buttons"]["layout"].lower()

        if keyboard_layout_name not in LAYOUTS:
            raise ValueError("Unsupported layout '%s'" % keyboard_layout_name)

        keyboard_layout = build_layout(LAYOUTS[keyboard_layout_name])

    mm_layout = build_layout(layout_multimedia)

    # -- Craft packet

    packet_length = len(buttons) * setting_info["button_field_length"]
    packet = [0x00] * packet_length

    for button, value in [(k.lower(), v) for k, v in mapping["buttons"].items()]:
        if button == "layout":
            continue
        if button not in buttons:
            raise ValueError("Unknown button name '%s'" % button)
        buttons[button]["value"] = value

    for button, data in buttons.items():
        value = data["default"]
        if "value" in data and data["value"].lower() != "default":
            value = data["value"]
        value = value.lower()

        if value in buttons:
            packet[data["offset"]] = buttons[value]["id"]
        elif value in button_special:
            packet[data["offset"]] = button_special[value]
        elif value in mm_layout:
            packet[data["offset"]] = button_multimedia
            packet[data["offset"] + 1] = mm_layout[value]
        elif value in keyboard_layout:
            packet[data["offset"]] = button_keyboard
            packet[data["offset"] + 1] = keyboard_layout[value]
        else:
            raise ValueError("Unknown button, key or action '%s'" % value)

    return packet


def cli_buttons_validator(setting_info):
    class CheckButtonsAction(argparse.Action):
        """Validate buttons from CLI"""

        def __call__(self, parser, namespace, value, option_string=None):
            if value.lower() == "default":
                setattr(namespace, self.dest.upper(), value)
                return

            if REGEXP_PARAM_STRING.match(value):
                is_valid, reason = is_buttons(value, setting_info)

                if is_valid:
                    setattr(namespace, self.dest.upper(), value)
                    return

                raise argparse.ArgumentError(self, "%s" % reason)

            else:
                raise argparse.ArgumentError(
                    self, "not a valid buttons mapping param: '%s'" % value
                )

    return CheckButtonsAction


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "buttons" type setting to the given CLI arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = "%s (default: %s)" % (
        setting_info["description"],
        str(setting_info["default"]).replace("%", "%%"),
    )
    cli_parser.add_argument(
        *setting_info["cli"],
        dest=setting_name,
        help=description,
        type=str,
        action=cli_buttons_validator(setting_info),
        metavar=setting_name.upper(),
    )



"""
This module contains varous helper functions.
"""


import re


#: A regual expression that matches the general form of a param string.
REGEXP_PARAM_STRING = re.compile(r"^\s*([a-zA-Z0-9_]+)\s*\(\s*(.+?)[\s;]*\)[\s;]*$")


def merge_bytes(*args):
    """Returns a single list of int from given int and list of int.

    :param int,list[int] args: Values to merge
    :rtype: list[int]

    >>> from rivalcfg.helpers import merge_bytes
    >>> merge_bytes(1, 2, 3)
    [1, 2, 3]
    >>> merge_bytes([1, 2], [3, 4])
    [1, 2, 3, 4]
    >>> merge_bytes(1, [2, 3], 4)
    [1, 2, 3, 4]
    """
    result = []
    for arg in args:
        if type(arg) in [list, tuple]:
            result.extend(arg)
        else:
            result.append(arg)
    return result


def module_ls(module):
    """List the content of the given Python module, ignoring private elements.

    :param module: The module to list.
    :rtype: [str]
    """
    return [e for e in dir(module) if not e.startswith("_")]


def parse_param_string(paramstr, value_parsers={}):
    """Parses a parameter string (used for rgbgrandiant).

    :param str paramstr: The parameter string.
    :param dict value_parsers: Additional parsers to parse values.
    :rtype: dict

    Example syntax::

       myparam(foo=1; bar=hello; baz=a, b, c)

    Once parsed, it will return this::

       {
           "myparam": {
               "foo": "1",
               "bar": "hello",
               "baz": "a, b, c",
           }
       }

    Additional parsers can be provided to parse the values, see examples
    bellow.

    >>> parse_param_string("hello(name=world)")
    {'hello': {'name': 'world'}}
    >>> parse_param_string("hello(name=world;)")
    {'hello': {'name': 'world'}}
    >>> parse_param_string("hello(name=world ) ;;")
    {'hello': {'name': 'world'}}
    >>> parse_param_string("foo(a=42; b=3.14)", value_parsers={
    ...     "foo": {
    ...         "a": int,
    ...         "b": float,
    ...     },
    ... })
    {'foo': {'a': 42, 'b': 3.14}}
    >>> parse_param_string("foobar[a=1]")
    Traceback (most recent call last):
        ...
    ValueError: invalid parameter string 'foobar[a=1]'
    """
    if not REGEXP_PARAM_STRING.match(paramstr):
        raise ValueError("invalid parameter string '%s'" % paramstr)

    name = REGEXP_PARAM_STRING.match(paramstr).group(1)
    params = REGEXP_PARAM_STRING.match(paramstr).group(2)

    try:
        result = {
            name: {
                k.strip(): v.strip()
                for k, v in [p.split("=") for p in params.split(";")]
            }
        }
    except ValueError:
        raise ValueError("invalid parameter string '%s'" % paramstr)

    if value_parsers:
        for key, value in result[name].items():
            if name in value_parsers and key in value_parsers[name]:
                result[name][key] = value_parsers[name][key](value)

    return result


def uint_to_little_endian_bytearray(number, size):
    """Converts an unsigned interger to a little endian bytearray.

    :param in number: The number to convert.
    :param in size: The length of the target bytearray.
    :rtype: [int]

    >>> uint_to_little_endian_bytearray(0x42, 1)
    [66]
    >>> uint_to_little_endian_bytearray(0x42, 2)
    [66, 0]
    >>> uint_to_little_endian_bytearray(0xFF42, 2)
    [66, 255]
    >>> uint_to_little_endian_bytearray(0xFF42, 4)
    [66, 255, 0, 0]
    >>> uint_to_little_endian_bytearray(0xFFFFFF, 2)
    Traceback (most recent call last):
        ...
    ValueError: integer overflow
    """
    if number > (2 ** (8 * size) - 1):
        raise ValueError("integer overflow")
    nle = [0] * size
    for i in range(size):
        nle[i] = number >> i * 8 & 0xFF
    return nle


def bytes_to_high_low_nibbles(byte):
    """Splits bytes into high and low nibbles.

    >>> bytes_to_high_low_nibbles(0xAB)
    (10, 11)
    """
    high = byte >> 4
    low = byte & 0x0F
    return high, low


def nibbles_to_byte(nibble1, nibble2):
    """Converts two nibbles into a single byte.

    >>> nibbles_to_byte(0xA, 0xB)
    171
    """
    return (nibble1 << 4) | nibble2
